/*************************************************************************\
*                                                                         *
*    File    : backoff_timer.v                                            *
*    Purpose : Backoff Timer Module                                       *
*    Author  : Novan Hartadi (novan@vlsi.itb.ac.id)                       *
*                                                                         *
/*************************************************************************\

/*************************************************************************\
*                                                                         *
*    Brief Description :                                                  *
*                                                                         *
*    The function of this module is to determine interval time for        *
*    backoff operation after a start_backoff is detected. This interval   *
*    time is calculated from multiplication between random number and     *
*    slot time, that is random x 128 clock cycles. This module takes      *
*    random number from Random Number Generator when start_backoff is     *
*    detected.                                                            *
*                                                                         *
\*************************************************************************/

`define SLOT_TIME (17'd128)
`define BACKOFF_IDLE (1'b0)
`define BACKOFF_RUN  (1'b1)

module backoff_timer (
  start_backoff,
  random,
  clk,
  reset_n,
  backoff_p
  );

input start_backoff;
input [9:0] random;
input clk;
input reset_n;
output backoff_p;

reg backoff_p;
reg [16:0] count_backoff;
reg [16:0] backoff_time;
reg state_backoff;

always @(posedge clk or negedge reset_n)
  if (!reset_n)
    state_backoff <= `BACKOFF_IDLE;
  else
    case (state_backoff)
      `BACKOFF_IDLE : if (start_backoff)
                        state_backoff <= `BACKOFF_RUN;
                      else
                        begin
                          state_backoff <= `BACKOFF_IDLE;
                          count_backoff <= 17'd0;
                          backoff_p <= 1'b0;
                        end
      `BACKOFF_RUN  : if (count_backoff == (backoff_time-1))
                        begin
                          state_backoff <= `BACKOFF_IDLE;
                          backoff_p <= 1'b1;
                        end
                      else
                        begin
                          state_backoff <= `BACKOFF_RUN;
                          count_backoff <= (count_backoff + 1'b1);
                        end
    endcase

always @(random)
  backoff_time = (random * `SLOT_TIME);

endmodule