/*************************************************************************\
*                                                                         *
*    File    : defer_counter.v                                            *
*    Purpose : Defer Counter Module                                       *
*    Author  : Novan Hartadi (novan@vlsi.itb.ac.id)                       *
*                                                                         *
/*************************************************************************\

/*************************************************************************\
*                                                                         *
*    Brief Description :                                                  *
*                                                                         *
*    The function of this module is to limit defer time for waiting       *
*    opportunity to transmit a packet of data. If transmit_new_p pulse    *
*    is asserted, that indicates a new packet is ready to be transmitted, *
*    then counter start counting. When the counter has reached the        *
*    maximum value of defer time, excess_deferral will be asserted until  *
*    transmit_new_p appears again. Thus, the defer counter will be        *
*    reseted when transmit_new_p is asserted and will stop counting when  *
*    transmit_available_p is asserted. The maximum value of defer time is *
*    2 times of the maximum length of Ethernet frame, that is 3036 byte   *
*    or 6072 clock cycles.                                                *
*                                                                         *
\*************************************************************************/

`define MAX_DEFER  (13'd6072)
`define DEFER_IDLE (1'b0)
`define DEFER_RUN  (1'b1)

module defer_counter (
  transmit_available_p, 
  transmit_new_p,
  clk, 
  reset_n, 
  excess_deferral
  );

input transmit_available_p;
input transmit_new_p;
input clk;
input reset_n;
output excess_deferral;

reg excess_deferral;
reg [12:0] count_defer;
reg state_defer;

always @(posedge clk or negedge reset_n)
  if (!reset_n)
    begin
      state_defer <= `DEFER_IDLE;
      excess_deferral <= 1'b0;
    end
  else
    case (state_defer)
      `DEFER_IDLE : if (transmit_new_p)
                      begin
                        excess_deferral <= 1'b0;
                        state_defer <= `DEFER_RUN;
                      end
                    else
                      begin
                        count_defer <= 13'd0;
                        state_defer <= `DEFER_IDLE;
                      end
      `DEFER_RUN  : if (count_defer == `MAX_DEFER-1)
                      begin
                        state_defer <= `DEFER_IDLE;
                        excess_deferral <= 1'b1;
                      end
                    else
                    if (transmit_available_p)
                      state_defer <= `DEFER_IDLE;
                    else
                      begin
                        count_defer <= (count_defer + 1'b1);
                        state_defer <= `DEFER_RUN;
                      end
    endcase    

endmodule