/*************************************************************************\
*                                                                         *
*    File    : frame_length_counter.v                                     *
*    Purpose : Frame Length Counter Module                                *
*    Author  : Novan Hartadi (novan@vlsi.itb.ac.id)                       *
*                                                                         *
/*************************************************************************\

/*************************************************************************\
*                                                                         *
*    Brief Description :                                                  *
*                                                                         *
*    The function of this module is to count the length of frame has been *
*    transmitted to MII.                                                  *
*    The length counter counts the length of frame while transmit_enable  *
*    is detected and will be reseted if transmit_enable is not asserted.  *
*    If transmit_data_end is detected, the FCS counter will start         *
*    counting.                                                            *
*    This module also detects excessive length and condition when data    *
*    has been transmitted is 64 byte. Note, that the length counter       *
*    counts Preamble and SFD at the beginning of operation, which makes   *
*    the maximum length of frame and 64 byte indicator increased by 16.   *
*                                                                         *
\*************************************************************************/

`define COUNT_LENGTH_IDLE (2'b00)
`define COUNT_LENGTH_RUN  (2'b01)
`define COUNT_LENGTH_FCS  (2'b11)
`define MIN_LENGTH        (12'd144)
`define MAX_LENGTH        (12'd3052)

module frame_length_counter (
  transmit_enable,
  transmit_fcs,
  clk,
  reset_n,
  count_length,
  count_fcs,
  excessive_length,
  transmit_64byte
  );

input transmit_enable;
input transmit_fcs;
input clk;
input reset_n;
output [11:0] count_length;
output [3:0] count_fcs;
output excessive_length;
output transmit_64byte;

reg [11:0] count_length;
reg [3:0] count_fcs;
reg excessive_length;
reg transmit_64byte;
reg [1:0] state_length;

function [11:0] inc_length;
input [11:0] count_length;
begin
  inc_length = count_length + 1'b1;
end
endfunction

always @(posedge clk or negedge reset_n)
  if (!reset_n)
    begin
      state_length <= `COUNT_LENGTH_IDLE;
      count_length <= 12'd0;
      count_fcs <=  4'd0;
    end
  else
    case (state_length)
      `COUNT_LENGTH_IDLE : if (transmit_enable)
                           begin
                             state_length <= `COUNT_LENGTH_RUN;
                             count_length <= 12'd1;
                           end
                           else
                             state_length <= `COUNT_LENGTH_IDLE;
      `COUNT_LENGTH_RUN  : if (!transmit_enable)
                             state_length <= `COUNT_LENGTH_IDLE;
                           else
                           if (transmit_fcs)
                             begin
                               state_length <= `COUNT_LENGTH_FCS;
                               count_length <= inc_length(count_length);
                               count_fcs <= 4'd1;
                             end
                           else
                             begin
                               state_length <= `COUNT_LENGTH_RUN;
                               count_length <= inc_length(count_length);
                             end
      `COUNT_LENGTH_FCS  : if (!transmit_enable)
                             state_length <= `COUNT_LENGTH_IDLE;
                           else
                             begin
                               state_length <= `COUNT_LENGTH_FCS;
                               count_length <= inc_length(count_length);
                               count_fcs <= (count_fcs + 1'b1);
                             end
    endcase

always @(count_length)
  begin
    transmit_64byte = (count_length > `MIN_LENGTH-1) ? 1'b1 : 1'b0;
    excessive_length = (count_length > `MAX_LENGTH-1) ? 1'b1 : 1'b0;
  end

endmodule