/*************************************************************************\
*                                                                         *
*    File    : ifg_timer.v                                                *
*    Purpose : Inter Frame Gap Timer Module                               *
*    Author  : Novan Hartadi (novan@vlsi.itb.ac.id)                       *
*                                                                         *
/*************************************************************************\

/*************************************************************************\
*                                                                         *
*    Brief Description :                                                  *
*                                                                         *
*    The function of this module is to assure time interval between two   *
*    packets of data or Inter Frame Gap (IFG) is 96 bit time or 24 clock  *
*    cycles. In half duplex mode, if counter can pass IFG without any     *
*    crs signal, transmit_available_p will be asserted. After asserting   *
*    transmit_available_p pulse, this module will monitor carrier signal  *
*    again. When operates in full duplex mode, this module will ignore    *
*    any crs signals and will count IFG after transmit_enable has been    *
*    disappeared.                                                         *
*    The IFG has two intervals time, IFG_1 for the first 60 bit times     *
*    and IFG_2 for the following 36 bit times of IFG. If crs signal       *
*    appears in the IFG_1 then the value of counter is reseted. But, if   *
*    the crs signal appears in the IFG_2 then the counter will keep       *
*    running until 96 bit times has been passed. This operation makes     *
*    every station has an equal opportunity to access the medium.         *
*                                                                         *
\*************************************************************************/

`define IFG_LENGTH_1 (5'd15)
`define IFG_LENGTH_2 (5'd22)
`define IFG_IDLE     (1'b0)
`define IFG_RUN      (1'b1)

module ifg_timer (
  crs,
  full_duplex,
  transmit_enable,
  clk,
  reset_n,
  transmit_available_p
  );

input crs;
input full_duplex;
input transmit_enable;
input clk;
input reset_n;
output transmit_available_p;

reg transmit_available_p;
reg [4:0] count_ifg;
reg state_ifg;

function [4:0] inc_ifg;
input [4:0] count_ifg;
begin
  inc_ifg = count_ifg + 1'b1;
end
endfunction

always @(posedge clk or negedge reset_n)
  if (!reset_n)
    state_ifg <= `IFG_IDLE;
  else
  if (full_duplex)
    case (state_ifg)
      `IFG_IDLE : if (!transmit_enable)
                    begin
                      state_ifg <= `IFG_RUN;
                      count_ifg <= 5'd1;
                      transmit_available_p <=  1'b0;
                    end
                  else
                    begin
                      state_ifg <= `IFG_IDLE;
                      count_ifg <= 5'd0;
                      transmit_available_p <= 1'b0;
                    end
      `IFG_RUN  : if (transmit_enable)
                    state_ifg <= `IFG_IDLE;
                  else
                  if (count_ifg == `IFG_LENGTH_2-1)
                    begin
                      state_ifg <= `IFG_IDLE;
                      transmit_available_p <= 1'b1;
                    end
                  else
                    begin
                      state_ifg <= `IFG_RUN;
                      count_ifg <= inc_ifg(count_ifg);
                    end
    endcase
  else
    case (state_ifg)
      `IFG_IDLE : if (!crs)
                    begin
                      state_ifg <= `IFG_RUN;
                      count_ifg <= 5'd1;
                      transmit_available_p <= 1'b0;
                    end
                  else
                    begin
                      state_ifg <= `IFG_IDLE;
                      count_ifg <= 5'd0;
                      transmit_available_p <= 1'b0;
                  end      
      `IFG_RUN  : if (crs && (count_ifg <= `IFG_LENGTH_1-1))
                    state_ifg <= `IFG_IDLE;
                  else
                  if (count_ifg == `IFG_LENGTH_2-1)
                    begin
                      state_ifg <= `IFG_IDLE;
                      transmit_available_p <= 1'b1;
                    end
                  else
                    begin
                      state_ifg <= `IFG_RUN;
                      count_ifg <= inc_ifg(count_ifg);
                    end
    endcase

endmodule